# Infinite stats for PicoScope 6, v0.1
# By lab!fyi <info@lab.fyi>
# http://lab.fyi/oscilloscope_scripts/infinite_stats/index.html
# This work is licensed under a Creative Commons Attribution-ShareAlike 4.0 International License.

import subprocess, sys, msvcrt, time

# CONTROL
# [Esc] - exit
# [Space] - reset

# CONFIG
rate = 1            # x=Loop_time_in_s; 0=Loop_rate_disabled
avg_source = 5      # 2=Value; 5=Average
round_input = 15    # decimal places
round_output = 6    # decimal places
round_hz = 1        # decimal places
show_header = True
pad_lines = 4       # pad display to X lines to avoid resizing window

unit_ratios = {
    'f': 1e-15,
    'p': 1e-12,
    'n': 1e-9,
    'µ': 1e-6,
    'm': 1e-3,
    'X': 1,
    'k': 1e3,
    'M': 1e6,
    'G': 1e9,
    'T': 1e12}

data = {}
old_data_keys = {}
new_data_keys = {}

def restart_line():
    sys.stdout.write('\r')
    sys.stdout.flush()

def parse_value(v):
    a = v.split(' ')
    if len(a[1]) == 0:
        a[1] = 'X'
    a[1] = a[1].replace('æ', 'µ')
    if (a[1][0] in unit_ratios):
        a[0] = round(float(a[0]) * unit_ratios[a[1][0]], round_input)
    else:
        a[0] = round(float(a[0]), round_input)
    return a

def unparse_value(v):
    if (v[1][0] in unit_ratios):
        return round(v[0] / unit_ratios[v[1][0]], round_output)
    else:
        return round(v[0], round_output)

last_update = time.time()
current_time = 0
def update_rate():
    global last_update, current_time
    current_time = time.time()
    if (last_update + rate) > current_time:
        time.sleep((last_update + rate) - current_time)
    current_time = time.time()
    hz = 1 / (current_time - last_update)
    last_update = current_time
    return hz

# main >>>

try:

    while True:

        cmd = ['picoscope_measurements.bat']
        output = subprocess.check_output(cmd, shell=False, universal_newlines=True)

        hz = round(update_rate(), round_hz) # rate control inside

        try:
            if msvcrt.kbhit():
                key = msvcrt.getch()
                if ord(key) == 32:
                    data = {}
                    new_data_keys= {}
                if ord(key) == 27:
                    break
        except KeyboardInterrupt:
            exit(0)
                
        old_data_keys = new_data_keys
        new_data_keys = {}
        index = 0

        nline = \
            "Index".rjust(10) + \
            "Ch".rjust(10) + \
            "f(Ch)".rjust(20) + \
            "Input".rjust(20) + \
            "Min".rjust(20) + \
            "Max".rjust(20) + \
            "Avg".rjust(20) + \
            "Count".rjust(10) + \
            "Hz".rjust(10)
        if show_header:
            print(nline)

        lines = output.splitlines()
        for line in lines:

            a = line.split(',')

            if len(a) == 10:
                index += 1            

                for i in range(2,7):
                    a[i] = parse_value(a[i])

                if not(a[0] in new_data_keys):
                    new_data_keys[a[0]] = []
                if not(a[1] in new_data_keys):
                    new_data_keys[a[0]].append(a[1])

                if not(a[0] in data): data[a[0]] = {}
                if not(a[1] in data[a[0]]):
                    data[a[0]][a[1]] = {'min': [a[3][0], '', 0], 'max': [a[4][0], '', 0], 'avg': [a[avg_source][0], '', 0], 'count': 1}
                else:
                    data[a[0]][a[1]]['count'] += 1

                # Min
                if a[3][0] < data[a[0]][a[1]]['min'][0]:
                    data[a[0]][a[1]]['min'][0] = a[3][0]
                data[a[0]][a[1]]['min'][1] = a[3][1] # Current unit

                # Max
                if a[4][0] > data[a[0]][a[1]]['max'][0]:
                    data[a[0]][a[1]]['max'][0] = a[4][0]
                data[a[0]][a[1]]['max'][1] = a[4][1] # Current unit
                
                # Exponential moving average
                data[a[0]][a[1]]['avg'][0] -= data[a[0]][a[1]]['avg'][0] / data[a[0]][a[1]]['count']
                data[a[0]][a[1]]['avg'][0] += a[avg_source][0] / data[a[0]][a[1]]['count']
                data[a[0]][a[1]]['avg'][1] = a[avg_source][1] # Current unit

                nline = \
                    str(index).rjust(10) + \
                    a[0].rjust(10) + \
                    a[1].rjust(20) + \
                    ('{:3.3f}'.format(unparse_value(a[avg_source])) + a[avg_source][1]).rjust(20) + \
                    ('{:3.3f}'.format(unparse_value(data[a[0]][a[1]]['min'])) + data[a[0]][a[1]]['min'][1]).rjust(20) + \
                    ('{:3.3f}'.format(unparse_value(data[a[0]][a[1]]['max'])) + data[a[0]][a[1]]['max'][1]).rjust(20) + \
                    ('{:3.3f}'.format(unparse_value(data[a[0]][a[1]]['avg'])) + data[a[0]][a[1]]['avg'][1]).rjust(20) + \
                    str(data[a[0]][a[1]]['count']).rjust(10) + \
                    str(hz).rjust(10)
                print(nline)

        data_keys_intersection = set(new_data_keys).intersection(set(old_data_keys))
        for key in (set(old_data_keys) - data_keys_intersection):
            data.pop(key)
        for key in set(new_data_keys):
            if key in set(old_data_keys):
                _data_keys_intersection = set(new_data_keys[key]).intersection(set(old_data_keys[key]))
                for _key in (set(old_data_keys[key]) - _data_keys_intersection):
                    data[key].pop(_key)

        for i in range(pad_lines - index):
            print()

except subprocess.CalledProcessError as e:
    sys.exit("'%s' failed, returned code %d" % (cmd, e.returncode))
except OSError as e:
    sys.exit("Failed to execute program '%s': '%s'" % (cmd, str(e)))
except KeyboardInterrupt:
    exit(0)

# <<< main
